(* Abstract Syntax Tree and functions for printing it *)

type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq |
          And | Or | Elemult | Elediv

type uop = Neg | Not

type typ = Int | Bool | Void | Char | String | Float | Matrix | Rangety

type import = Import of string

type expr =
    Literal of int
  | BoolLit of bool
  | CharLit of char
  | StringLit of string
  | FloatLit of string
  | MatrixLit of float array * (int * int)
  | MatAccess of expr * expr * expr 
  | MatSlice of expr * expr * expr
  | StringAccess of expr * expr
  | Id of string
  | Binop of expr * op * expr
  | Unop of uop * expr
  | Assign of expr * expr
  | Call of string * expr list
  | CastExpr of typ * expr
  | Noexpr
  | Range of index * index
  and index = Beg | End | Ind of expr

type bind_decl = typ * string (* (typ, string) *)

type bind_init = bind_decl * expr (* ((typ, string), expr) *)

type bind =
   Decl of bind_decl
 | DeclInit of bind_init

type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | While of expr * stmt
  | Continue
  | Break
  | VarDecl of bind

type func_decl = {
    typ : typ;
    fname : string;
    formals : bind_decl list;
    body : stmt list;
  }

type program = import list * bind list * func_decl list

(* Pretty-printing functions *)

let string_of_op = function
    Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"
  | Elemult -> ".*"
  | Elediv -> "./"

let string_of_uop = function
    Neg -> "-"
  | Not -> "!"

let string_of_typ = function
    Int -> "int"
  | Bool -> "bool"
  | Void -> "void"
  | Char -> "char"
  | String -> "str"
  | Float -> "float"
  | Matrix -> "matrix"
  | Rangety -> "range"

let rec string_of_matrix scalar cols _ =
  "[ " ^ (let rec string_of_elems = function
      [] -> ""
    | h::t -> string_of_expr h ^
              (if (List.length t) == 0 then ""
              else if (List.length t) mod cols == 0 then "; "
              else ", ") ^
              string_of_elems t
  in string_of_elems scalar) ^ " ]"

and string_of_expr = function
    Literal(l) -> string_of_int l
  | BoolLit(true) -> "true"
  | BoolLit(false) -> "false"
  | CharLit(c) -> "'" ^ String.escaped (String.make 1 c) ^ "'"
  | StringLit(s) -> "\"" ^ String.escaped s ^ "\""
  | FloatLit(f) -> f
  | MatSlice(m, b, c) -> string_of_expr m ^ "[" ^ string_of_expr b ^ "," ^ string_of_expr c ^ "]"
  | MatrixLit(m,(a,b)) -> 
      let arr = (List.map (fun x -> FloatLit(Printf.sprintf "%g" x))) (Array.to_list m) in string_of_matrix arr a b
  | MatAccess(m, b, c) -> "Accessing " ^ string_of_expr m ^" element [" ^string_of_expr b^ "," ^ string_of_expr c ^ "]"
  | Range(s, e) -> 
      let a = (match s with 
          Beg -> "Beg" 
        | End -> "End" 
        | Ind(e) -> string_of_expr e) 
      and b = (match e with 
          Beg -> "Beg" 
        | End -> "End" 
        | Ind(e) -> string_of_expr e) 
      in a ^ " : " ^ b
  | StringAccess(s, i) -> string_of_expr s ^ "[" ^ string_of_expr i ^ "]"
  | Id(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | Unop(o, e) -> string_of_uop o ^ string_of_expr e
  | Assign(v, e) -> string_of_expr v ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | CastExpr(t, e) -> "(" ^ string_of_typ t ^ ") " ^ string_of_expr e
  | Noexpr -> ""

let string_of_vdecl = function
    Decl(decl) -> string_of_typ (fst decl) ^ " " ^ (snd decl) ^ ";\n"
  | DeclInit(decl, exp) -> string_of_typ (fst decl) ^ " " ^ (snd decl) ^ " = " ^ string_of_expr exp ^ ";\n"

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s
  | Continue -> "continue;\n"
  | Break -> "break;\n"
  | VarDecl(v) -> string_of_vdecl v



  let string_of_fdecl fdecl =
  string_of_typ fdecl.typ ^ " " ^
  fdecl.fname ^ "(" ^ String.concat ", " (List.map snd fdecl.formals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
