(* Semantic checking for the Yamml compiler *)

open Ast
open Sast

module StringHash = Hashtbl.Make(
  struct type t = string (* type of keys *) 
  let equal x y = x = y (* use structural comparison *) 
  let hash = Hashtbl.hash (* generic hash function *) 
end);;


module StringMap = Map.Make(String)

(* Semantic checking of the AST. Returns an SAST if successful,
   throws an exception if something is wrong.

   Check each global variable, then check each function *)

let check (globals, functions) =

  let report_duplicate exceptf list =
    let rec helper = function
	      n1 :: n2 :: _ when n1 = n2 -> raise (Failure (exceptf n1))
      | _ :: t -> helper t
      | [] -> ()
    in helper (List.sort compare list)
  in

  (* Raise an exception if a given binding is to a void type *)
  let check_not_void exceptf = function
      (Void, n) -> raise (Failure (exceptf n))
    | _ -> ()
  in

  (* Verify a list of bindings has no void types or duplicate names *)
  
  let check_binds (kind : string) (bd : bind_decl list) = 
    List.iter (check_not_void (fun n -> "illegal void "^kind^ " " ^n)) bd;
    report_duplicate (fun n -> "duplicate "^ kind ^" " ^ n) (List.map snd bd);

  in

  let rec global_symbols = function 
    [] -> []
    | hd::tl -> match hd with 
        DeclInit(d) -> fst d::global_symbols tl
      | Decl(d) -> d::global_symbols tl
  in 
  
  (**** Check global variables ****)

  ignore(check_binds "global" (global_symbols globals));

  (**** Check functions ****)

  (* Collect function declarations for built-in functions: no bodies *)
  let built_in_decls = 
    let add_bind map (name, ty, rty) = StringMap.add name {
      typ = rty;
      fname = name; 
      formals = List.fold_left (fun l t -> l @ [(t, "x")]) [] ty;
      body = [] } map
    in List.fold_left add_bind StringMap.empty [ 
      ("print", [Int], Void);
      ("printf", [Float], Void);
      ("printb", [Bool], Void);
      ("printc", [Char], Void); 
      ("printm", [Matrix], Int);
      ("printbig", [Int], Void);
      ("prints", [String], Void); 
      ("height", [Matrix], Int); 
      ("width", [Matrix], Int);
      ("sum", [Matrix], Float); 
      ("mean", [Matrix], Float); 
      ("trans", [Matrix], Matrix);
      ("empty", [Int; Int], Matrix);
      (* OpenCV functions *)
      ("filter2D", [Matrix; Matrix; Int; Matrix], Void);
      ("imread", [String], Matrix);
      ("imwrite", [String; Matrix], Void);
      ("invert", [Matrix; Matrix], Void);
       ]
  in

  (* Add function name to symbol table *)
  let add_func map fd = 
    let built_in_err = "function " ^ fd.fname ^ " may not be defined"
    and dup_err = "duplicate function " ^ fd.fname
    and make_err er = raise (Failure er)
    and n = fd.fname (* Name of the function *)
    in match fd with (* No duplicate functions or redefinitions of built-ins *)
         _ when StringMap.mem n built_in_decls -> make_err built_in_err
       | _ when StringMap.mem n map -> make_err dup_err  
       | _ ->  StringMap.add n fd map 
  in

  (* Collect all function names into one symbol table *)
  let function_decls = List.fold_left add_func built_in_decls functions
  in
  
  (* Return a function from our symbol table *)
  let find_func s = 
    try StringMap.find s function_decls
    with Not_found -> raise (Failure ("unrecognized function " ^ s))
  in

  let _ = find_func "main" in (* Ensure "main" is defined *)


  (* Return a variable from our local symbol table *)
  let type_of_identifier s map =
    try StringMap.find s map
    with Not_found -> raise (Failure ("undeclared identifier " ^ s))
  in

  (* Raise an exception if the given rvalue type cannot be assigned to
      the given lvalue type *)
  let check_assign lvaluet rvaluet err =
      if lvaluet = rvaluet then lvaluet else raise (Failure err)
  in   

  (* Return a semantically-checked expression, i.e., with a type *)
  let rec expr e map = match e with
        Literal  l -> (Int, SLiteral l)
      | FloatLit l -> (Float, SFloatLit l)
      | StringLit l -> (String, SStringLit l)
      | CharLit l -> (Char, SCharLit l)
      | BoolLit l  -> (Bool, SBoolLit l)
      | MatrixLit(l,(r,c)) -> 
          if Array.length l = r * c 
          then (Matrix, SMatrixLit(l, (r,c)))
          else raise ( Failure ("illegal Matrix Dimension"))
      | Noexpr     -> (Void, SNoexpr)
      | Id s       -> (type_of_identifier s map, SId s)
      | Assign(var, e) as ex -> 
          let lt = (match var with
              Id s -> type_of_identifier s map
            | MatAccess(_,_,_) -> Float
            | StringAccess(_,_) -> Char
            | _ -> raise (Failure "illegal assignment LHS"))
          and (rt, e') = expr e map in
          let err = "illegal assignment " ^ string_of_typ lt ^ " = " ^ 
            string_of_typ rt ^ " in " ^ string_of_expr ex
          in (check_assign lt rt err, SAssign(expr var map, (rt, e')))
      | Unop(op, e) as ex -> 
          let (t, e') = expr e map in
          let ty = match op with
            Neg when t = Int || t = Float -> t
          | Not when t = Bool -> Bool
          | _ -> raise (Failure ("illegal unary operator " ^ 
                                 string_of_uop op ^ string_of_typ t ^
                                 " in " ^ string_of_expr ex))
          in (ty, SUnop(op, (t, e')))
      | Binop(e1, op, e2) as e -> 
          let (t1, _) = expr e1 map
          and (t2, _) = expr e2 map in
          (* All binary operators require operands of the same type *)
          let same = t1 = t2 in
          (* Determine expression type based on operator and operand types *)
          let ty = match op with
            Add | Sub | Mult | Div when same && t1 = Int   -> Int
          | Add | Sub | Mult | Div when same && t1 = Float -> Float
          | Elemult | Elediv | Mult when same && t1 = Matrix -> Matrix
          | Add | Sub | Mult | Div when t1 = Float && t2 = Int -> Float
          | Add | Sub | Mult | Div when t1 = Int && t2 = Float -> Float
          | Equal | Neq            when same               -> Bool
          | Less | Leq | Greater | Geq
                     when same && (t1 = Int || t1 = Float) -> Bool
          | And | Or when same && t1 = Bool -> Bool
          | Div when t1 = Matrix && t2 == Float -> Matrix
          | Mult when t1 = Matrix && t2 == Float -> Matrix
          | Mult when t2 = Matrix && t1 == Float -> Matrix
          | _ -> raise (
	      Failure ("illegal binary operator " ^
                       string_of_typ t1 ^ " " ^ string_of_op op ^ " " ^
                       string_of_typ t2 ^ " in " ^ string_of_expr e))
      in
          let (t1, e1') = match e1 with 
            Literal l when ty = Float -> (Float, SFloatLit (string_of_int l))
          | _ -> expr e1 map

          and (t2, e2') =  match e2 with 
            Literal l when ty = Float -> (Float, SFloatLit (string_of_int l))
          | _ -> expr e2 map


          in (ty, SBinop((t1, e1'), op, (t2, e2')))

      | MatAccess(em, e1, e2) -> let (tm, sem) = (expr em map) in
        (match tm with Matrix ->
          let (t1, se1) = expr e1 map 
          in  let (t2,se2) = expr e2 map 
              in (match t1,t2 with
                  Int,Int -> (Float, SMatAccess((tm,sem), (t1, se1), (t2, se2)))
                | _ -> raise ( Failure ("Index is not a integer!")))
          | _ -> raise ( Failure ("matrix access on non-matrix"))
        )

      | MatSlice(m, r1, r2) -> 
        let r1', r2' = expr r1 map, expr r2 map in
        let m' = expr m map in
        (Matrix, SMatSlice(m', r1', r2'))
      | Range(r1, r2) -> (match r1, r2 with 
              Beg, End -> (Rangety, SRange(SBeg, SEnd))
            | Beg, Ind(i) -> (Rangety, SRange(SBeg, SInd(expr i map)))
            | Ind(i), End -> (Rangety, SRange(SInd(expr i map), SEnd))
            | Ind(i1), Ind(i2) when i1 <= i2 -> (Rangety, SRange(SInd(expr i1 map), SInd(expr i2 map)))
            | _ -> raise (Failure "Invalid ranges")
          )
      | CastExpr(ty, e) -> let se = expr e map in (ty, SCastExpr(ty, se))
      | StringAccess(s, i) ->
          let (ts, ses) = expr s map in
          (match ts with String ->
            let (t2,se2) = expr i map in
            (match t2 with
                Int -> (Char, SStringAccess((ts, ses), (t2, se2)))
              | _ -> raise ( Failure ("Index is not an integer!")))
            | _ -> raise ( Failure ("string access on non-string")))
      | Call(fname, args) as call -> 
          let fd = find_func fname in
          let param_length = List.length fd.formals in
          if List.length args != param_length then
            raise (Failure ("expecting " ^ string_of_int param_length ^ 
                            " arguments in " ^ string_of_expr call))
          else let check_call (ft, _) e = 
            let (et, e') = expr e map in 
            let err = "illegal argument found " ^ string_of_typ et ^
              " expected " ^ string_of_typ ft ^ " in " ^ string_of_expr e
            in (check_assign ft et err, e')
          in 
          let args' = List.map2 check_call fd.formals args
          in (fd.typ, SCall(fname, args'))
    in


  let check_function func =
    (* Make sure no formals or locals are void or duplicates *)
    
    check_binds "formal" func.formals;
    (* check_binds "local" func.locals; *)
   
    
    (* Build local symbol table of variables for this function *)
    let symbols = List.fold_left (fun m (ty, name) -> StringMap.add name ty m)
	                StringMap.empty (global_symbols globals @ func.formals)

    in
  
    let check_bool_expr e vars = 
      let (t', e') = expr e vars 
      and err = "expected Boolean expression in " ^ string_of_expr e
      in if t' != Bool then raise (Failure err) else (t', e') 
    in

    (* Return a semantically-checked statement i.e. containing sexprs *)
    let rec check_stmt stmt vars = match stmt with
        Expr e -> (SExpr (expr e vars), vars)
      | If(p, b1, b2) -> (SIf(check_bool_expr p vars, fst (check_stmt b1 vars), fst (check_stmt b2 vars)), vars)
      | For(e1, e2, e3, st) ->
	        (SFor(expr e1 vars, check_bool_expr e2 vars, expr e3 vars, fst(check_stmt st vars)), vars)
      | While(p, s) -> (SWhile(check_bool_expr p vars, fst (check_stmt s vars)), vars)
      | Return e -> let (t, e') = expr e vars in
          if t = func.typ then (SReturn (t, e'), vars)
          else raise (
          Failure ("return gives " ^ string_of_typ t ^ " expected " ^
          string_of_typ func.typ ^ " in " ^ string_of_expr e))
      | VarDecl(v) -> (match v with
          Decl(ty, n) -> (SVarDecl(SDecl(ty, n)), StringMap.add n ty vars)
        | DeclInit((ty, n), e) -> (SVarDecl(SDeclInit((ty, n), expr e vars)), StringMap.add n ty vars))
	    
	    (* A block is correct if each statement is correct and nothing
	       follows any Return statement.  Nested blocks are flattened. *)
      | Block sl -> 
          let bvars = StringMap.empty in
          let rec check_stmt_list stmt_list vars bvars =
            let check s ss bvars =
              let checked = check_stmt s vars in
              let checked_rest = check_stmt_list ss (snd checked) bvars in
              (fst checked :: fst checked_rest, snd checked_rest)
            in
            match stmt_list with
              [Return _ as s] -> ([fst (check_stmt s vars)], bvars)
            | Return _ :: _   -> raise (Failure "nothing may follow a return")
            | VarDecl v as s :: ss ->
              let check_decl ty n = match ty with
                Void -> raise (Failure ("illegal void local variable " ^ n))
              | _ -> (match StringMap.find_opt n bvars with
                  Some _ -> raise (Failure ("duplicate local variable " ^ n))
                | None -> ())
              in (match v with
                    Decl(ty, n) -> ignore(check_decl ty n); check s ss (StringMap.add n ty bvars) (* (fst checked, StringMap.add n ty (snd checked)) *)
                  | DeclInit((ty, n), _) -> ignore(check_decl ty n); check s ss (StringMap.add n ty bvars)) (* (fst checked, StringMap.add n ty (snd checked))) *)
            | s :: ss         -> check s ss bvars
            | []              -> ([], bvars)
          in (SBlock(fst (check_stmt_list sl vars bvars)), vars)
      | Continue -> (SContinue, vars)
      | Break -> (SBreak, vars)

    in (* body of check_function *)
    { styp = func.typ;
      sfname = func.fname;
      sformals = func.formals;
      sbody = match fst (check_stmt (Block func.body) symbols) with
        SBlock(sl) -> sl
      | _ -> raise (Failure ("internal error: block didn't become a block?"))
    }
  in 

  let check_global glob =
    (* Build local symbol table of variables for this function *)
    let symbols = List.fold_left (fun m (ty, name) -> StringMap.add name ty m)
                  StringMap.empty (global_symbols globals)
                  (* concats two diff types into a list *)

    in
    match glob with
        DeclInit(b, e) -> SDeclInit(b, expr e symbols)
      | Decl(d) -> SDecl(d)
  in (List.map check_global globals, List.map check_function functions)
